BossPanel_Quests = BossPanelPlugin:new({
	name          = BossPanel_QuestsLocals.NAME,
	description   = BossPanel_QuestsLocals.DESCRIPTION,
	version       = "0.1.2",
	releaseDate   = "03-25-2006",
	aceCompatible = 103,
	bpCompatible  = 900,
	author        = "Kemayo",
	email         = "kemayo@gmail.com",
	website       = "http://kemayo.wowinterface.com/",
	category      = "quests",
	db            = AceDatabase:new("BossPanel_QuestsDB"),
	defaults      = {
		showLevels = TRUE,
		showDifficulty = TRUE,
		showArea = FALSE,
		showCurrentAreaOnly = FALSE,
		showCurrentAreaDescriptionOnly = FALSE,
		showDescription = TRUE,
	},
	cmd           = AceChatCmd:new(BossPanel_QuestsLocals.COMMANDS, BossPanel_QuestsLocals.CMD_OPTIONS),
	loc           = BossPanel_QuestsLocals,
	hasIcon       = TRUE,
	canHideText   = TRUE,
	overrideTooltip = TRUE,
	
	-- Methods
	ShowTooltip = function(self)
		self:OpenChildFrame(BossPanel_QuestsFrame)
		self.idleTime = 0
	end,
	
	Frame_OnUpdate = function(self)
		if not self.enteredFrame then
			self.idleTime = self.idleTime + arg1
			if self.idleTime >= 2 then
				BossPanel_QuestsFrame:Hide()
			end
		end
	end,
	
	Frame_OnEnter = function(self)
		BossPanel_QuestsFrame:Show()
	end,
	
	Frame_OnLeave = function(self)
		self.enteredFrame = FALSE
	end,
	
	Quest_Click = function(self)
		self.idleTime = 0
		if this:GetID()>20 then
			self.quests[this:GetID()-20].Hide = not self.quests[this:GetID()-20].Hide
			self:UpdateTooltip()
		else
			local id = this:GetID()
			local questIndex
			for i, zone in self.quests do
				if not zone.Hide then
					for ii, quest in zone.quests do
						id = id-1
						if id==0 then
							questIndex = quest.id
						end
					end
				end
			end
			
			if IsShiftKeyDown() then			
				-- Shift-click toggles quest-watch on this quest.
				if ( IsQuestWatched(questIndex) ) then
					RemoveQuestWatch(questIndex);
					QuestWatch_Update();
				else
					-- Set error if no objectives
					if ( GetNumQuestLeaderBoards(questIndex) == 0 ) then
						UIErrorsFrame:AddMessage(QUEST_WATCH_NO_OBJECTIVES, 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
						return;
					end
					-- Set an error message if trying to show too many quests
					if ( GetNumQuestWatches() >= MAX_WATCHABLE_QUESTS ) then
						UIErrorsFrame:AddMessage(format(QUEST_WATCH_TOO_MANY, MAX_WATCHABLE_QUESTS), 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
						return;
					end
					AddQuestWatch(questIndex);
					QuestWatch_Update();
				end
			else
				if QuestLogFrame:IsVisible() then
					if self.lastIndex == questIndex then
						HideUIPanel(QuestLogFrame)
					end
				else
					ShowUIPanel(QuestLogFrame)
				end		
				QuestLog_SetSelection(questIndex)
				self.lastIndex = questIndex
			end
		end		
	end,
	
	IsShowingLevels = function(self)
		return self.data.showLevels
	end,
	
	ToggleShowingLevels = function(self, silent)
		self.data.showLevels = not self.data.showLevels
		--self:Update()
		return self.data.showLevels
	end,
	
	IsShowingDifficulty = function(self)
		return self.data.showDifficulty
	end,
	
	ToggleShowingDifficulty = function(self, silent)
		self.data.showDifficulty = not self.data.showDifficulty
		--self:Update()
		return self.data.showDifficulty
	end,
	
	IsShowingArea = function(self)
		return self.data.showArea
	end,
	
	ToggleShowingArea = function(self, silent)
		self.data.showArea = not self.data.showArea
		--self:Update()
		return self.data.showArea
	end,
	
	IsShowingCurrentAreaOnly = function(self)
		return self.data.showCurrentAreaOnly
	end,
	
	ToggleShowingCurrentAreaOnly = function(self, silent)
		self.data.showCurrentAreaOnly = not self.data.showCurrentAreaOnly
		--self:Update()
		return self.data.showCurrentAreaOnly
	end,
	
	IsShowingCurrentAreaDescriptionOnly = function(self)
		return self.data.showCurrentAreaDescriptionOnly
	end,
	
	ToggleShowingCurrentAreaDescriptionOnly = function(self, silent)
		self.data.showCurrentAreaDescriptionOnly = not self.data.showCurrentAreaDescriptionOnly
		--self:Update()
		return self.data.showCurrentAreaDescriptionOnly
	end,
	
	IsShowingDescription = function(self)
		return self.data.showDescription
	end,
	
	ToggleShowingDescription = function(self, silent)
		self.data.showDescription = not self.data.showDescription
		--self:Update()
		return self.data.showDescription
	end,
	
	Initialize = function(self)
		--self:Update()
		self.numTooltipItems = 0
		self.numQuests = 0
		self.quests = {}
		BossPanel_QuestsFrameText:SetText(BossPanel_QuestsLocals.TEXT_QUESTS)
		BossPanel_Quest1:SetPoint("TOPLEFT", 5, -30)
		local prev, button = BossPanel_Quest1
		for i=2, 30 do
			button = getglobal("BossPanel_Quest"..i)
			button:SetPoint("TOPLEFT", prev, "BOTTOMLEFT")
			prev = button
		end
	end,
	
	Enable = function(self)
		self:Update()
		self:RegisterEvent("QUEST_LOG_UPDATE", "Update")
		self:RegisterEvent("PLAYER_ENTERING_WORLD", "OnPlayerEnteringWorld")
		self:RegisterEvent("PLAYER_LEAVING_WORLD", "OnPlayerLeavingWorld")
	end,
	
	Disable = function(self)
		self:UnregisterEvent("QUEST_LOG_UPDATE")
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("PLAYER_LEAVING_WORLD")
	end,
	
	Report = function(self)
		self.cmd:report({
			{text=self.loc.ARGUMENT_LEVELS, val=(self:IsShowingLevels() and 1 or 0), map=BossPanelLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_DIFFICULTY, val=(self:IsShowingDifficulty() and 1 or 0), map=BossPanelLocals.MAP_ONOFF},
			{text=self.loc.ARGUMENT_AREA, val=(self:IsShowingArea() and 1 or 0), map=BossPanelLocals.MAP_ONOFF},
		})
	end,
	
	MenuSettings = function(self)
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_LEVELS,
			func = function()
				self:ToggleShowingLevels()
			end,
			checked = self:IsShowingLevels(),
			keepShownOnClick = TRUE,
		})
		
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_DIFFICULTY,
			func = function()
				self:ToggleShowingDifficulty()
			end,
			checked = self:IsShowingDifficulty(),
			keepShownOnClick = TRUE,
		})
		
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_AREA,
			func = function()
				self:ToggleShowingArea()
			end,
			checked = self:IsShowingArea(),
			keepShownOnClick = TRUE,
		})
		
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_CURRENT_AREA_ONLY,
			func = function()
				self:ToggleShowingCurrentAreaOnly()
			end,
			checked = self:IsShowingCurrentAreaOnly(),
			keepShownOnClick = TRUE,
		})
		
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_DESCRIPTION,
			func = function()
				self:ToggleShowingDescription()
			end,
			checked = self:IsShowingDescription(),
			keepShownOnClick = TRUE,
		})
		
		UIDropDownMenu_AddButton({
			text = self.loc.MENU_SHOW_CURRENT_AREA_DESCRIPTION_ONLY,
			func = function()
				self:ToggleShowingCurrentAreaDescriptionOnly()
			end,
			checked = self:IsShowingCurrentAreaDescriptionOnly(),
			keepShownOnClick = TRUE,
		})
	end,
	
	OnPlayerEnteringWorld = function(self)
		self:RegisterEvent("QUEST_LOG_UPDATE", "Update")
	end,
	
	OnPlayerLeavingWorld = function(self)
		self:UnregisterEvent("QUEST_LOG_UPDATE")
	end,
	
	UpdateData = function(self)
		--UpdateText and UpdateTooltip both need to know things about the questlog.  Better to do this once.
		local startingQuestLogSelection = GetQuestLogSelection()
		local numQuests = GetNumQuestLogEntries()
		local quests = {}
		local zoneIndex = 0
		local quest
		
		for quest = 1, numQuests, 1 do
			SelectQuestLogEntry(quest)
			local strQuestLogTitleText, strQuestLevel, strQuestTag, isHeader, isCollapsed, isComplete = GetQuestLogTitle(quest)
			local strQuestDescription, strQuestObjectives = GetQuestLogQuestText()
			local q = {}
			local thisQuest = "  "
			local strQuestLevelTag = ""
			
			if strQuestTag == self.loc.ELITE then
				strQuestLevelTag = "+"
			elseif strQuestTag == self.loc.DUNGEON then
				strQuestLevelTag = "D"
			elseif strQuestTag == self.loc.PVP then
				strQuestLevelTag = "P"
			elseif strQuestTag == self.loc.PVP then
				strQuestTag = "R"
			end
			
			if isHeader then
				zoneIndex = zoneIndex + 1
				quests[zoneIndex] = { ["zone"]=strQuestLogTitleText, ["quests"]={} }
			else
				if isComplete then
					isComplete = format("(%s)",self.loc.QUEST_DONE)
				elseif isComplete == -1 then
					isComplete = format("(%s)",self.loc.QUEST_FAILED)
				else
					isComplete = ""
				end
				
				q['title'] = strQuestLogTitleText
				q['level'] = strQuestLevel
				q['tag'] = strQuestLevelTag
				q['complete'] = isComplete
				q['objective'] = strQuestObjectives
				
				local lb = {}
				if (GetNumQuestLeaderBoards() > 0) then
					for ii=1, GetNumQuestLeaderBoards(), 1 do
						local desc, qtype, done = GetQuestLogLeaderBoard(ii)
						
						if done then
							done = format("(%s)",self.loc.QUEST_DONE)
						else
							done = ""
						end
						
						lb[ii] = {['description']=desc, ['done']=done}
					end
				end
				q['leaderboard'] = lb
				q['id'] = quest
				
				table.insert(quests[zoneIndex]['quests'],q)
			end
		end
		
		self.quests = quests
		self.numQuests = numQuests - zoneIndex
		
		self:HowManyLines()
		
		SelectQuestLogEntry(startingQuestLogSelection)
	end,
	
	UpdateText = function(self)
		local maxQuests = 20
		local color = BossPanelUtils.GetThresholdHexColor((maxQuests - self.numQuests) / maxQuests)
		
		self:SetText(BossPanelUtils.Colorize(color,format("%d/%d", self.numQuests, maxQuests)))
	end,
	
	UpdateTooltip = function(self)
		local currentZone = GetZoneText()
		local showArea = self:IsShowingArea()
		local showDescription = self:IsShowingDescription()
		local showCurrentAreaOnly = self:IsShowingCurrentAreaOnly()
		local showCurrentAreaDescriptionOnly = self:IsShowingCurrentAreaDescriptionOnly()
		local qi, qc, height = 1, 0, 35
		
		-- To remove when WoW hits 1.10
--[[		if self:HowManyLines() > 39 then
			-- Trim down the lines shown.
			if not showCurrentAreaDescriptionOnly then showCurrentAreaDescriptionOnly = TRUE
			elseif showDescription then showDescription = FALSE
			elseif not showCurrentAreaOnly then showCurrentAreaOnly = True end
		end]]
		
		-- Add a blank line, because otherwise the header seems to overwrite the first area.
--		BossPanelTooltip:AddBlankLine(nil)
		for i,zone in self.quests do
			if showArea and ((currentZone == zone['zone']) or not showCurrentAreaOnly) then
				local button = getglobal("BossPanel_Quest"..qi)
				getglobal("BossPanel_Quest"..qi.."Right"):SetText("")
				if zone.Hide then
					getglobal("BossPanel_Quest"..qi.."Left"):SetText("  "..BossPanelUtils.Colorize("bfffbf",zone['zone']))				
					getglobal("BossPanel_Quest"..qi.."Check"):Show()
				else
					getglobal("BossPanel_Quest"..qi.."Left"):SetText(BossPanelUtils.Colorize("bfffbf",zone['zone']))				
					getglobal("BossPanel_Quest"..qi.."Check"):Hide()
				end
				button:SetWidth(350)
				getglobal("BossPanel_Quest"..qi.."Left"):SetWidth(275)
				local butHeight = getglobal("BossPanel_Quest"..qi.."Left"):GetHeight()+4				
				button:SetHeight(butHeight)
				button:Show()
				qc = qc+1
				button:SetID(20+qc)
				qi = qi+1
				height = height + butHeight
--				BossPanelTooltip:AddLine(nil, BossPanelUtils.Colorize("bfffbf",zone['zone']))
			end
			if not zone.Hide then
				for ii,quest in zone['quests'] do
					if (currentZone == zone['zone']) or not showCurrentAreaOnly then
						local thisQuest
						local questLevel = ""					
						
						if self:IsShowingLevels() then
							questLevel = format("[%s%s] ", quest['level'], quest['tag'])
						end
						
						thisQuest = questLevel..quest['title']
						
						if self:IsShowingDifficulty() then
							local thisQuestColor = GetDifficultyColor(quest['level'])
							thisQuest = format("|c%02X%02X%02X%02X%s|r", 255, thisQuestColor.r * 255, thisQuestColor.g * 255, thisQuestColor.b * 255, thisQuest)
						end
						
						local left, right = "  "..thisQuest, quest['complete']
--					BossPanelTooltip:AddDoubleLine(nil, "  "..thisQuest, quest['complete'])
						
						if showDescription and ((currentZone == zone['zone']) or not showCurrentAreaDescriptionOnly) then
							if table.getn(quest['leaderboard']) > 0 then
								for iii,goal in quest['leaderboard'] do
									local goalColor = self:GetGoalColor(goal['description'])
									left = left.."\n    "..BossPanelUtils.Colorize(goalColor,goal['description'])
									right = right.."\n"..BossPanelUtils.Colorize(goalColor,goal['done'])
--								BossPanelTooltip:AddDoubleLine(nil, "    "..BossPanelUtils.Colorize(goalColor,goal['description']), BossPanelUtils.Colorize(goalColor,goal['done']), 1,1,1,1,1,1, BossPanelTooltip:GetNormalFontSize()-1)
								end
							else
								left, right = left.."\n    "..BossPanelUtils.Colorize("7f7f7f",quest['objective']), right.."\n"
--							BossPanelTooltip:AddLine(nil, "    "..BossPanelUtils.Colorize("7f7f7f",quest['objective']), 1, 1, 1, TRUE, "LEFT", BossPanelTooltip:GetNormalFontSize()-1)
							end
						end
						local button = getglobal("BossPanel_Quest"..qi)
						getglobal("BossPanel_Quest"..qi.."Left"):SetText(left)
						getglobal("BossPanel_Quest"..qi.."Right"):SetText(right)
						if ( IsQuestWatched(quest.id) ) then
							getglobal("BossPanel_Quest"..qi.."Check"):Show()
						else
							getglobal("BossPanel_Quest"..qi.."Check"):Hide()
						end
						button:SetWidth(350)
						getglobal("BossPanel_Quest"..qi.."Left"):SetWidth(275)
						local butHeight = getglobal("BossPanel_Quest"..qi.."Left"):GetHeight()+4
						button:SetHeight(butHeight)
						button:Show()
						button:SetID(qi-qc)
						qi = qi+1
						height = height + butHeight
					end
				end
			end
		end
		
		for i = qi,30 do
			getglobal("BossPanel_Quest"..i):Hide()
		end
		BossPanel_QuestsFrame:SetWidth(350)
		BossPanel_QuestsFrame:SetHeight(height)
	end,
	
	HowManyLines = function(self)
		-- Tooltip can (in 1.9) only be 40 lines long.  How many lines will this be?
		-- Note, this is annoyingly clunky.  It's a copy of the logic for UpdateTooltip.
		-- I'm looking forward to 1.10, when I can remove this.
		-- Starts at 1, as a blank line is added currently.
		local numLines = 1
		local currentZone = GetZoneText()
		
		for i,zone in self.quests do
			if self:IsShowingArea() and ((currentZone == zone['zone']) or not self:IsShowingCurrentAreaOnly()) then
				numLines = numLines + 1
			end
			for ii,quest in zone['quests'] do
				if (currentZone == zone['zone']) or not self:IsShowingCurrentAreaOnly() then
					numLines = numLines + 1
					if self:IsShowingDescription() and ((currentZone == zone['zone']) or not self:IsShowingCurrentAreaDescriptionOnly()) then
						if table.getn(quest['leaderboard']) > 0 then
							for iii,goal in quest['leaderboard'] do
								numLines = numLines + 1
							end
						else
							numLines = numLines + 1
						end
					end
				end
			end
		end
		self.numLines = numLines
		return numLines
	end,
	
	OnClick = function(self)
		ToggleQuestLog()
	end,
	
	GetGoalColor = function(self, goalText)
		local i, j, itemName, numItems, numNeeded = string.find(goalText, "(.*):%s*([-%d]+)%s*/%s*([-%d]+)%s*$");
		if numItems ~= nil then
			-- This quest involves items!
			return BossPanelUtils.GetThresholdHexColor(numItems / numNeeded)
		else
			-- This quest doesn't involve items.  Let's leave it grey.  We're lazy.
			return "7f7f7f"
		end
	end,
})

BossPanel_Quests:RegisterForLoad()